"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib.framework.injectable import Injectable
from rapid.workflow.action_dal import ActionDal


class ActionInstanceService(Injectable):
    __injectables__ = {'action_dal': ActionDal}

    def __init__(self, action_dal):
        """
        :param action_dal:
        :type action_dal: ActionDal
        """
        self.action_dal = action_dal

    def finish_action_instance(self, _id, post_data):
        self.action_dal.complete_action_instance(_id, post_data)

    def get_action_instance_by_id(self, _id):
        return self.action_dal.get_action_instance_by_id(_id)

    def edit_action_instance(self, _id, changes):
        return self.action_dal.partial_edit(_id, changes)

    def reset_action_instance(self, _id, complete_reset=False, check_status=False):
        return self.action_dal.reset_action_instance(_id, complete_reset, check_status)

    def reset_pipeline_instance(self, pipeline_instance_id):
        return self.action_dal.reset_pipeline_instance(pipeline_instance_id)

    def callback(self, _id, post_data):
        return self.action_dal.callback_action_instance(_id, post_data)

    def reconcile_pipeline_instances(self):
        return self.action_dal.reconcile_pipeline_instances()



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except
import logging
import datetime

try:
    import simplejson as out_json
except ImportError:
    import json as out_json


from flask import request
from flask.wrappers import Response
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.sql.expression import asc

from rapid.lib.exceptions import InvalidObjectException
from rapid.lib.store_service import StoreService
from rapid.workflow.data.models import PipelineEvent
from rapid.lib import api_key_required, get_db_session
from rapid.lib.constants import StatusConstants, ModuleConstants
from rapid.lib.exceptions import VcsNotFoundException
from rapid.lib.framework.injectable import Injectable
from rapid.lib.modules import CiModule
from rapid.master.data.database.dal.general_dal import GeneralDal
from rapid.workflow.data.models import Action, Pipeline, Stage, Workflow, PipelineInstance, PipelineParameters


logger = logging.getLogger("rapid")


class PipelineDal(GeneralDal, Injectable):
    __injectables__ = {ModuleConstants.CI_MODULE: CiModule, 'flask_app': None, 'queue_constants': None}

    def __init__(self, ci_module, queue_constants, flask_app=None):
        """

        :type ci_module: :class:`rapid.lib.modules.modules.CiModule`
        :type queue_constants: QueueHandlerConstants
        :param flask_app: Flask
        :return:
        """
        super(PipelineDal, self).__init__()
        self.app = flask_app
        self.ci_module = ci_module
        self.queue_constants = queue_constants

    def is_serviceable(self, model):
        return model == Pipeline

    def register_url_rules(self, flask_app):
        self.app = flask_app
        flask_app.add_url_rule('/api/pipelines/create', 'create_pipeline', api_key_required(self.create_pipeline), methods=['POST'])
        flask_app.add_url_rule('/api/pipelines/<int:pipeline_id>/start', 'start_pipeline_instance', api_key_required(self.start_pipeline_instance), methods=['POST'])

    def create_pipeline(self):
        json = request.get_json()
        return Response(self._get_pipeline(json), content_type='application/json')

    def start_pipeline_instance_via_reponame(self, repo, json_data=None):
        """
        Start a pipeline_instance identified by the the repo name
        :param repo: VCS repo name or Vcs.repo
        :param json_data: dict for the Data in the PipelineInstance
        :return:
        :rtype PipelineInstance
        """
        vcs = self.ci_module.get_vcs_by_repo_name(repo)
        if vcs is not None:
            if vcs.pipeline_id is not None:
                return self.create_pipeline_instance(vcs.pipeline_id, json_data, vcs_id=vcs.id)
            if vcs.active:
                raise VcsNotFoundException("The repo[{}] did not have a default pipeline defined.".format(repo))
        raise VcsNotFoundException("The repo [{}] is not found in the system".format(repo))

    def start_pipeline_instances_via_pipeline_id(self, pipeline_id, json_data=None):
        for session in get_db_session():
            return self.create_pipeline_instance(pipeline_id, json_data, in_session=session)

    def start_pipeline_instance(self, pipeline_id):
        data = request.get_json()
        return "It worked!" if self.create_pipeline_instance(pipeline_id, data) else "It Failed!"

    def get_pipeline_by_id(self, pipeline_id, session=None):
        if session is None:
            for db_session in get_db_session():
                return db_session.query(Pipeline).get(pipeline_id).serialize()
        else:
            return session.query(Pipeline).get(pipeline_id)

    def get_pipeline_events_by_pipeline_id(self, pipeline_id, session=None):
        if session is None:
            for db_session in get_db_session():
                return [event.serialize() for event in db_session.query(PipelineEvent).filter(PipelineEvent.pipeline_id == pipeline_id).all()]
        else:
            return session.query(PipelineEvent).filter(PipelineEvent.pipeline_id == pipeline_id).all()

    def get_pipeline_instance_by_id(self, pipeline_instance_id, session=None):
        if session is None:
            for db_session in get_db_session():
                return db_session.query(PipelineInstance).get(pipeline_instance_id).serialize()
        else:
            return session.query(PipelineInstance).get(pipeline_instance_id)

    def _get_pipeline(self, json):
        if json:
            for session in get_db_session():
                pipeline = Pipeline(**{"name": json['name'], 'active': json['active']})
                session.add(pipeline)
                session.flush()
                pipeline.stages.extend(self.get_stages(json, pipeline, session))

                response = out_json.dumps(pipeline.serialize())
                try:
                    session.commit()
                finally:
                    if session:
                        session.expunge_all()
                        session.close()
                        session = None

                return response

        raise BaseException("No pipeline was created.")

    def get_actions(self, json, pipeline, workflow, session):
        actions = []
        if 'actions' in json:
            for tmp_action in json['actions']:
                action = Action(**tmp_action)
                action.order = len(actions)
                action.pipeline_id = pipeline.id
                action.workflow_id = workflow.id
                session.add(action)
                actions.append(action)
        return actions

    def get_workflows(self, json, pipeline, stage, session):
        workflows = []
        if 'workflows' in json:
            for tmp_workflow in json['workflows']:
                workflow = Workflow(
                    **{"name": tmp_workflow['name'], 'active': tmp_workflow['active'], "order": len(workflows),
                       "stage_id": stage.id})
                session.add(workflow)
                session.flush()
                workflow.actions.extend(self.get_actions(tmp_workflow, pipeline, workflow, session))
                workflows.append(workflow)
        return workflows

    def get_stages(self, json, pipeline, session):
        stages = []
        if 'stages' in json:
            for tmp_stage in json['stages']:
                stage = Stage(**{"name": tmp_stage['name'], "active": tmp_stage['active'], "order": len(stages),
                                 "pipeline_id": pipeline.id})
                session.add(stage)
                session.flush()
                stage.workflows.extend(self.get_workflows(tmp_stage, pipeline, stage, session))
                stages.append(stage)
        return stages

    def new_alchemy_encoder(self):
        _visited_objs = []

        class AlchemyEncoder(out_json.JSONEncoder):
            def default(self, obj):  # pylint: disable=arguments-differ,method-hidden
                if isinstance(obj.__class__, DeclarativeMeta):
                    # don't re-visit self
                    if obj in _visited_objs:
                        return None
                    _visited_objs.append(obj)

                    # an SQLAlchemy class
                    fields = {}
                    for field in [x for x in dir(obj) if
                                  not x.startswith('_') and x != 'metadata' and x != 'query' and x != 'query_class']:
                        fields[field] = obj.__getattribute__(field)
                    # a json-encodable dict
                    return fields

                return out_json.JSONEncoder.default(self, obj)

        return AlchemyEncoder

    def get_actions_query(self, session, pipeline_id):
        return session.query(Stage, Workflow, Action) \
            .filter(Stage.pipeline_id == pipeline_id) \
            .filter(Stage.id == Workflow.stage_id) \
            .filter(Workflow.id == Action.workflow_id) \
            .order_by(asc(Stage.order)) \
            .order_by(asc(Workflow.id)) \
            .order_by(asc(Action.order))

    def create_pipeline_instance(self, pipeline_id, json_data=None, vcs_id=None, in_session=None):
        if in_session:
            return self._process_pipeline(pipeline_id, vcs_id, json_data, in_session)
        for session in get_db_session():
            return self._process_pipeline(pipeline_id, vcs_id, json_data, session)

    def _process_pipeline(self, pipeline_id, vcs_id, json_data, session):
        pipeline = session.query(Pipeline).get(pipeline_id)
        if pipeline is not None and pipeline.active:
            pipeline_instance = PipelineInstance(pipeline_id=pipeline_id, status_id=StatusConstants.INPROGRESS,
                                                 created_date=datetime.datetime.utcnow(),
                                                 start_date=datetime.datetime.utcnow())
            session.add(pipeline_instance)
            session.flush()
            self._setup_pipeline(session, pipeline_id, pipeline_instance.id)
            try:
                if json_data and 'parameters' in json_data:
                    for parameter, value in json_data['parameters'].items():
                        tmp = PipelineParameters(parameter=parameter, value=value)
                        tmp.pipeline_instance_id = pipeline_instance.id
                        session.add(tmp)

                        if parameter == "commit":
                            # Look for vcs and get the ID
                            if vcs_id is None:
                                vcs = self.ci_module.get_vcs_by_pipeline_id(pipeline_id, session=session)
                                vcs_id = vcs.id if vcs is not None else None
                            else:
                                self.ci_module.create_git_commit(value, vcs_id, pipeline_instance_id=pipeline_instance.id, session=session)

            except Exception as exception:
                logger.error("Creating Pipeline Instance failed.")
                logger.error(exception)

            session.commit()
            create_pipeline_instance = pipeline_instance.serialize()
            return create_pipeline_instance
        try:
            logger.info("Inactive pipeline: {}".format(pipeline.name))
        except AttributeError:
            pass
        return {"message": "Invalid Pipeline, or inactive pipeline"}

    def cancel_pipeline_instance(self, pipeline_instance_id):
        for session in get_db_session():
            pipeline_instance = self.get_pipeline_instance_by_id(pipeline_instance_id, session)
            if pipeline_instance:
                for action_instance in pipeline_instance.action_instances:
                    for client in StoreService.get_clients(self.app).values():
                        if action_instance.status_id <= StatusConstants.SUCCESS and client.get_uri() == action_instance.assigned_to:
                            self.queue_constants.cancel_worker(action_instance.serialize())
                pipeline_instance.status_id = StatusConstants.CANCELED
                pipeline_instance.end_date = datetime.datetime.utcnow()
                session.commit()
            else:
                raise InvalidObjectException("Pipeline Instance not found", 404)
        return {"message": "Running clients have been canceled and pipeline canceled."}

    def _setup_pipeline(self, session, pipeline_id, pipeline_instance_id):
        current_stage = None
        current_workflow = None
        first_action_instance = True
        added_objects = []
        for (stage, workflow, action) in self.get_actions_query(session, pipeline_id):
            if current_stage is None:
                current_stage = stage.convert_to_instance()
                current_stage.pipeline_instance_id = pipeline_instance_id
                session.add(current_stage)
                session.flush()
                added_objects.append(current_stage)
            elif current_stage.stage_id != stage.id:
                break

            if current_workflow is None or current_workflow.workflow_id != workflow.id:
                first_action_instance = True
                current_workflow = workflow.convert_to_instance()
                current_workflow.stage_instance_id = current_stage.id
                session.add(current_workflow)
                session.flush()
                added_objects.append(current_workflow)

            slices = max(1, action.slices)

            for slice_num in range(slices):
                action_instance = action.convert_to_instance()
                action_instance.workflow_instance_id = current_workflow.id
                action_instance.pipeline_instance_id = pipeline_instance_id
                action_instance.slice = "{}/{}".format(slice_num + 1, slices)

                if first_action_instance:
                    action_instance.status_id = StatusConstants.READY

                session.add(action_instance)
                added_objects.append(action_instance)

            session.flush()

            if first_action_instance:
                first_action_instance = False
        return added_objects



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import abc

from rapid.lib.constants import Constants


class AbstractParser(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, workspace='', failures_only=False, failures_count=False):
        self.workspace = workspace
        self.failures_only = failures_only
        self.failures_count = failures_count

    @abc.abstractmethod
    def _parse_lines(self, lines):
        yield

    @staticmethod
    @abc.abstractmethod
    def get_type():
        yield

    def prepare_summary(self):
        return {Constants.STATUS_FAILED: 0, Constants.STATUS_SUCCESS: 0, Constants.STATUS_SKIPPED: 0, Constants.FAILURES_COUNT: self.failures_count}

    def parse(self, lines, ignore_type_check=False):
        if self.get_type() != lines[0].strip() and not ignore_type_check:
            raise Exception("Invalid first line identifier")

        if ignore_type_check:
            return self._parse_lines(lines)
        
        return self._parse_lines(lines[1:])



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

app_configuration

Revision ID: 9327e407fdfd
Revises: 8d5f3b5854d0
Create Date: 2020-11-16 11:31:47.643156

"""

# revision identifiers, used by Alembic.
from rapid.workflow.data.models import AppConfiguration

revision = '9327e407fdfd'
down_revision = '8d5f3b5854d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('app_configurations',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('process_queue', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_app_configurations_id'), 'app_configurations', ['id'], unique=False)
    op.bulk_insert(AppConfiguration.__table__, [
        {'id': 1, 'process_queue': True}
    ])


def downgrade():
    op.drop_index(op.f('ix_app_configurations_id'), table_name='app_configurations')
    op.drop_table('app_configurations')



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from rapid.lib.framework.injectable import Injectable
from rapid.workflow.events.event_dal import EventDal


class EventService(Injectable):
    __injectables__ = {'event_dal': EventDal}

    def __init__(self, event_dal=None):
        """
        :param event_dal:
        :type event_dal: workflow.events.EventDal.EventDal
        """
        self.event_dal = event_dal

    def trigger_possible_event(self, pipeline_instance, action_instance, session=None):
        return self.event_dal.trigger_possible_event(pipeline_instance, action_instance, session)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from rapid.lib.constants import EventTypes
from rapid.lib.framework.injectable import Injectable
from rapid.master.data.database.dal.general_dal import GeneralDal
from rapid.workflow.data.dal.pipeline_dal import PipelineDal
from rapid.workflow.events.event_handler import EventHandlerFactory


class EventDal(GeneralDal, Injectable):
    __injectables__ = {'pipeline_dal': PipelineDal}

    def __init__(self, pipeline_dal):
        """
        :param pipeline_dal:
        :type pipeline_dal: PipelineDal
        """
        self.pipeline_dal = pipeline_dal

    def trigger_possible_event(self, pipeline_instance, action_instance, session=None):
        """
        :param pipeline_instance:
        :type pipeline_instance: rapid.workflow.data.models.PipelineInstance
        :param action_instance:
        :type action_instance: rapid.workflow.data.models.ActionInstance
        :param session:
        :type session:
        :return:
        :rtype:
        """
        for event in self.pipeline_dal.get_pipeline_events_by_pipeline_id(pipeline_instance.pipeline_id, session=session):
            try:
                handler = EventHandlerFactory.get_event_handler(EventTypes(event.event_type_id))
                handler.handle_event(pipeline_instance, action_instance, event)
            except Exception:  # pylint: disable=broad-except
                pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=no-member,too-few-public-methods

import datetime

from sqlalchemy.orm import relationship, backref
from sqlalchemy import Column, String, ForeignKey, Integer, DateTime, Boolean, Text, Enum

from rapid.lib import get_declarative_base
from rapid.master.data.database.models.base.base_model import BaseModel
from rapid.lib.constants import VcsReleaseStepType
Base = get_declarative_base()


class Release(BaseModel, Base):
    name = Column(String(255), nullable=False, index=True)
    date_created = Column(DateTime(), nullable=False, default=datetime.datetime.utcnow, index=True)
    status_id = Column(Integer, ForeignKey('statuses.id'), nullable=False, index=True)
    commit_id = Column(Integer, ForeignKey('commits.id'), nullable=False, index=True)
    integration_id = Column(Integer, ForeignKey('integrations.id'), index=True)

    status = relationship('Status')
    integration = relationship('Integration')
    commit = relationship('Commit', backref=backref('release', uselist=False))


class StepIntegration(BaseModel, Base):
    step_id = Column(Integer, ForeignKey('steps.id'), nullable=False, index=True)
    integration_id = Column(Integer, ForeignKey('integrations.id'), nullable=False, index=True)


class Step(BaseModel, Base):
    name = Column(String(100), nullable=False)
    custom_id = Column(String(25), nullable=False)
    status_id = Column(Integer, ForeignKey('statuses.id'), nullable=False, index=True)
    user_required = Column(Boolean, default=False, nullable=False)
    release_id = Column(Integer, ForeignKey('releases.id'), nullable=False, index=True)
    sort_order = Column(Integer, default=0)

    release = relationship("Release", lazy='subquery', backref="steps")
    status = relationship('Status')
    integrations = relationship("Integration", secondary="step_integrations")


class StepUser(BaseModel, Base):
    step_id = Column(Integer, ForeignKey('steps.id'), nullable=False, index=True)
    user_id = Column(Integer, ForeignKey('users.id'), nullable=False, index=True)
    date_created = Column(DateTime(), nullable=False, default=datetime.datetime.utcnow)


class StepUserComment(BaseModel, Base):
    step_user_id = Column(Integer, ForeignKey('step_users.id'), nullable=False)
    comment = Column(Text)


class User(BaseModel, Base):
    name = Column(String(150), nullable=False)
    username = Column(String(150), nullable=False)
    active = Column(Boolean, default=True, nullable=False)


class VcsRelease(BaseModel, Base):
    search_filter = Column(String(500), nullable=False)
    notification_id = Column(String(250), nullable=False)
    vcs_id = Column(Integer, ForeignKey('vcs.id'), nullable=False, index=True)
    auto_release = Column(Boolean, nullable=False, default=False)

    vcs = relationship('Vcs', lazy='subquery', backref='product_release')
    steps = relationship("VcsReleaseStep", backref='vcs_release')


class VcsReleaseStep(BaseModel, Base):
    name = Column(String(250), nullable=False)
    custom_id = Column(String(250), nullable=False)
    user_required = Column(Boolean, default=False, nullable=False)
    sort_order = Column(Integer, default=0)
    type = Column(Enum(*list(map(lambda x: x.name, VcsReleaseStepType))), nullable=False, default='PRE')
    vcs_release_id = Column(Integer, ForeignKey('vcs_releases.id'), nullable=False, index=True)


__all__ = ['Release', 'StepIntegration', 'Step', 'StepUser', 'StepUserComment', 'StepIntegration', 'User', 'VcsRelease', 'VcsReleaseStep']



from abc import ABCMeta, abstractproperty

from rapid.lib.work_request import WorkRequest
from rapid.workflow.queue_handlers.queue_handler import QueueHandler


class ContainerHandler(QueueHandler):
    __metaclass__ = ABCMeta

    @abstractproperty
    def container_identifier(self):
        yield

    def can_process_work_request(self, work_request):
        # type: (WorkRequest) -> bool
        try:
            (grain_type, image) = self._get_grain_type_split(work_request.grain)
            return grain_type == self.container_identifier
        except ValueError:
            pass
        return False

    def can_process_action_instance(self, action_instance):
        # type: (dict) -> bool
        try:
            (grain_type, image) = self._get_grain_type_split(action_instance['grain'])
            return grain_type == self.container_identifier
        except (ValueError, KeyError):
            pass
        return False



# Always prefer setuptools over distutils
"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import ssl
import sys
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

ssl._create_default_https_context = ssl._create_unverified_context

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

try:
    execfile('./rapid/lib/version.py')
except NameError:
    exec(open('./rapid/lib/version.py').read())

requirements = [
    'flask',
    'requests==2.25.0',
    'futures',
    'jsonpickle',
    'simplejson==3.10.0',
    'simpleeval'
]

if sys.version_info[0] == 2:
    requirements.append('enum34==1.0.4')

setup(
        name='rapid-framework',
        version=__version__,
        description='Rapid Framework',
        long_description=long_description,

        # The project's main homepage.
        url='https://github.com/BambooHR/rapid',

        # Author details
        author='Michael Bright',
        author_email='mbright@bamboohr.com',

        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            # Indicate who your project is intended for
            'Topic :: Software Development :: Build Tools',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 2.7'
        ],

        # What does your project relate to?
        keywords='internal ci jenkinskiller',

        # You can just specify the packages manually here if your project is
        # simple. Or you can use find_packages().
        packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

        # List run-time dependencies here.  These will be installed by pip when
        # your project is installed. For an analysis of "install_requires" vs pip's
        # requirements files see:
        # https://packaging.python.org/en/latest/requirements.html
        install_requires=requirements,

        # List additional groups of dependencies here (e.g. development
        # dependencies). You can install these using the following syntax,
        # for example:
        # $ pip install -e .[dev,test]
        extras_require={
            'test': ['coverage', 'nose', 'ddt', 'mock', 'MagicMock'],
            'master': ['alembic', 'SQLAlchemy>=1.0.6', 'Flask-SQLAlchemy>=2.3.0', 'mysqlclient', 'pygithub', 'simpleeval'],
            'windows-client': ['waitress==1.4.4'],
            'client': []
        },

        # To provide executable scripts, use entry points in preference to the
        # "scripts" keyword. Entry points provide cross-platform support and allow
        # pip to create the appropriate form of executable for the target platform.
        entry_points={
            'console_scripts': [
                'rapid=rapid.__main__:main',
            ],
        },
)




"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from flask.globals import request

from rapid.ci.services.vcs_service import VcsService
from rapid.lib import api_key_required, json_response
from rapid.lib.framework.injectable import Injectable


class CIController(Injectable):
    __injectables__ = {'vcs_service': VcsService, 'flask_app': None}

    def __init__(self, vcs_service, flask_app):
        """

        :type vcs_service: :class:`rapid.ci.services.vcs_service.VcsService`
        :type flask_app: :class:`flask.Flask`
        :return:
        """
        self.vcs_service = vcs_service
        self.flask_app = flask_app

    def register_url_rules(self):
        self.flask_app.add_url_rule("/api/commits/<path:commit_identifier>/version", "create_version_for_commit", api_key_required(self.create_version_for_commit), methods=["PUT"])

    @json_response()
    def create_version_for_commit(self, commit_identifier):
        return self.vcs_service.create_version_for_commit(commit_identifier, request.get_json()['version'])



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.ci.ci_controller import CIController
from rapid.ci.services.vcs_service import VcsService
from rapid.ci.vcs.github_controller import GithubController
from rapid.lib.framework.ioc import IOC


def register_ioc_globals(flask_app):  # pylint: disable=unused-argument
    IOC.register_global('ci_module', IOC.get_class_instance(VcsService))


def configure_module(flask_app):
    load_model_layers()

    try:
        IOC.get_class_instance(GithubController, flask_app).register_url_rules(flask_app)
        IOC.get_class_instance(CIController).register_url_rules()
    except Exception: # pylint: disable=broad-except
        import traceback
        traceback.print_exc()


def load_model_layers():
    # Used for JIT loading dependencies.
    import rapid.ci.data.models  # pylint: disable=unused-import





"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import time
import requests

try:
    import simplejson as json
except ImportError:
    import json

from rapid.lib import api_key_required
from ...lib.base_controller import BaseController


class Main(BaseController):

    def __init__(self):
        self.flask_app = None

    def register_url_rules(self, flask_app):
        flask_app.add_url_rule('/fire_registration', 'registration', api_key_required(self.fire_registration))
        self.flask_app = flask_app

    def fire_registration(self):
        request = requests.post('http://rapidci.local/client/register',
                                data=json.dumps({'grains': {'something': 'more'}}),
                                headers={'content-type': 'application/json',
                                         'X-RapidCI-Port': '8002',
                                         'X-RapidCI-Time': str(time.time() * 1000)},
                                verify=self.flask_app.rapid_config.ignore_cert_verify)
        if request.status_code == 200:
            return "Success!"

        raise BaseException("There was a problem registrating.")



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import uuid
from ..lib.configuration import Configuration

# pylint: disable=too-many-instance-attributes,too-few-public-methods


class MasterConfiguration(Configuration):
    def __init__(self, file_name=None):
        self.port = None
        self.queue_time = None
        self.db_connect_string = None
        self.data_type = None
        self.queue_manager = None
        self.queue_consider_late_time = None
        self.register_api_key = None
        self.api_key = None
        self.github_user = None
        self.github_webhooks_key = None
        self.github_default_parameters = None
        self.install_uri = None
        self.install_options = None
        self.verify_certs = None
        self.custom_reports_dir = None
        self.static_file_directory = None
        self.basic_auth_user = None
        self.basic_auth_pass = None
        self.basic_auth = None
        self.ecs_config_file = None

        super(MasterConfiguration, self).__init__(file_name)

    def set_basic_auth(self, basic_auth):
        if basic_auth:
            (self.basic_auth_user, self.basic_auth_pass) = basic_auth.split(':')

    def _set_values(self, parser):
        super(MasterConfiguration, self)._set_values(parser)
        self.set_basic_auth(self.basic_auth)

    @property
    def section_mapping(self):
        return {
            'master': {
                'port': [8080, int],
                'queue_time': [6, int],
                'db_connect_string': ['sqlite:///data.db'],
                'data_type': ['inmemory'],
                'queue_manager': [True, bool],
                'queue_consider_late_time': [10, float],
                'register_api_key': [str(uuid.uuid3(uuid.NAMESPACE_OID, 'RegisterApiKey!')).replace('-', '')],
                'api_key': [str(uuid.uuid3(uuid.NAMESPACE_OID, 'MasterApiKey!')).replace('-', '')],
                'github_user': [None],
                'github_pass': [None],
                'github_webhooks_key': [None],
                'github_default_parameters': [None],
                'custom_reports_dir': [None],
                'static_file_directory': [None],
                'basic_auth': [None]
            },
            'general': {
                'install_uri': ['https://pypi.python.org/pypi/'],
                'install_options': [''],
                'verify_certs': [True, bool],
                'log_file': [None],
                'extensions': [None, list, ',']
            },
            'ecs': {
                'ecs_config_file': [None]
            }
        }



from flask import Flask

from rapid.client import ClientCommunicator
from rapid.client.executor import Executor
from rapid.lib.work_request import WorkRequest
from rapid.client.client_configuration import ClientConfiguration


class ActionInstanceRunner(object):

    def __init__(self, rapid_config):
        # type: (ClientConfiguration) -> None
        self.client_config = rapid_config

    def run_action_instance(self, app, action_instance_id):
        # type: (Flask, int) -> None
        communicator = ClientCommunicator(self.client_config.master_uri, self.client_config.quarantine_directory, app, self.client_config.verify_certs)
        self.client_config.is_single_use = True
        
        communicator.register(self.client_config)
        request_json = communicator.get_work_request_by_action_instance_id(action_instance_id)
        executor = Executor(WorkRequest(request_json),
                            self.client_config.master_uri,
                            workspace=self.client_config.workspace,
                            quarantine=self.client_config.quarantine_directory,
                            verify_certs=self.client_config.verify_certs,
                            rapid_config=self.client_config)
        executor.verify_work_request()
        executor.start(False)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=singleton-comparison,broad-except
import logging
import datetime
import random

import time

from sqlalchemy import func, and_, asc
from sqlalchemy.orm import joinedload, aliased
from sqlalchemy.sql.expression import exists

from rapid.lib.exceptions import InvalidObjectException
from rapid.lib.store_service import StoreService
from rapid.lib.constants import StatusConstants, StatusTypes, ModuleConstants
from rapid.lib.work_request import WorkRequest
from rapid.lib.framework.injectable import Injectable
from rapid.lib.modules import QaModule
from rapid.lib import get_db_session
from rapid.workflow.workflow_engine import InstanceWorkflowEngine
from rapid.workflow.data.dal.status_dal import StatusDal
from rapid.workflow.data.models import ActionInstance, PipelineInstance, PipelineParameters, Status, \
    PipelineStatistics, Statistics, StageInstance, WorkflowInstance, ActionInstanceConfig, AppConfiguration
from rapid.master.data.database.dal.general_dal import GeneralDal
from rapid.workflow.event_service import EventService

logger = logging.getLogger('rapid')


class ActionDal(GeneralDal, Injectable):
    __injectables__ = {ModuleConstants.QA_MODULE: QaModule,
                       'store_service': StoreService,
                       'event_service': EventService,
                       'flask_app': None,
                       'status_dal': StatusDal,
                       'queue_constants': None}
    last_sent = None

    def __init__(self, qa_module=None, store_service=None, event_service=None, flask_app=None, status_dal=None, queue_constants=None):
        """

        :param qa_module: rapid.master.modules.modules.QaModule
        :type  store_service: StoreService
        :type event_service: EventService
        :type status_dal: StatusDal
        :return:
        """
        self.qa_module = qa_module
        self.store_service = store_service
        self.event_service = event_service
        self.flask_app = flask_app
        self.status_dal = status_dal
        self.queue_constants = queue_constants

    def _get_ready_work_requests(self, session, work_requests, results):
        """
        :param session:
        :type session: sqlalchemy.orm.scoping.scoped_session
        :param work_requests:
        :type work_requests: dict
        :param results:
        :type results: list
        :return:
        :rtype:
        """
        for action_instance, pipeline_parameters, action_instance_config in session.query(ActionInstance, PipelineParameters, ActionInstanceConfig) \
                .outerjoin(PipelineParameters, PipelineParameters.pipeline_instance_id == ActionInstance.pipeline_instance_id) \
                .outerjoin(ActionInstanceConfig, ActionInstanceConfig.action_instance_id == ActionInstance.id) \
                .filter(ActionInstance.status_id == StatusConstants.READY) \
                .filter(ActionInstance.manual == 0) \
                .filter(ActionInstance.pipeline_instance_id == PipelineInstance.id) \
                .filter(PipelineInstance.status_id == StatusConstants.INPROGRESS) \
                .order_by(PipelineInstance.priority.desc(),
                          PipelineInstance.created_date.asc(),
                          PipelineInstance.id.asc(),
                          ActionInstance.order.asc(),
                          ActionInstance.slice.asc()).all():
            self.configure_work_request(action_instance, pipeline_parameters, work_requests, results)

    def _get_stalled_work_requests(self, session, work_requests, results):
        """
        :param session:
        :type session:
        :param work_requests:
        :type work_requests: dict
        :param results:
        :type results: list
        :return:
        :rtype:
        """
        action_alias = aliased(ActionInstance)
        inner_query = ~exists()\
            .where(action_alias.workflow_instance_id == WorkflowInstance.id) \
            .where(action_alias.order < ActionInstance.order)\
            .where(action_alias.end_date == None)\
            .correlate(ActionInstance) \
            .correlate(WorkflowInstance)
        for action_instance, pipeline_parameters, action_instance_config in session.query(ActionInstance, PipelineParameters, ActionInstanceConfig) \
                .join(PipelineInstance, PipelineInstance.id == ActionInstance.pipeline_instance_id) \
                .join(WorkflowInstance, WorkflowInstance.id == ActionInstance.workflow_instance_id) \
                .outerjoin(PipelineParameters, PipelineParameters.pipeline_instance_id == ActionInstance.pipeline_instance_id) \
                .outerjoin(ActionInstanceConfig, ActionInstanceConfig.action_instance_id == ActionInstance.id) \
                .filter(ActionInstance.status_id == StatusConstants.NEW) \
                .filter(ActionInstance.manual == 0) \
                .filter(PipelineInstance.status_id == StatusConstants.INPROGRESS) \
                .filter(inner_query) \
                .order_by(PipelineInstance.priority.desc(),
                          PipelineInstance.created_date.asc(),
                          PipelineInstance.id.asc(),
                          ActionInstance.order.asc(),
                          ActionInstance.slice.asc()).all():
            action_instance.configuration = action_instance_config
            self.configure_work_request(action_instance, pipeline_parameters, work_requests, results)

    def configure_work_request(self, action_instance, pipeline_parameters, work_requests, results, include_configuration=True):
        """
        :param action_instance:
        :type action_instance: ActionInstance
        :param pipeline_parameters:
        :type pipeline_parameters: PipelineParameters
        :param results:
        :type results: list
        :return:
        :rtype:
        """
        work_request = None
        if action_instance.id not in work_requests:
            work_request = WorkRequest(action_instance.serialize())
            work_request.action_instance_id = action_instance.id
            work_request.pipeline_instance_id = action_instance.pipeline_instance_id
            work_request.workflow_instance_id = action_instance.workflow_instance_id
            work_request.slice = action_instance.slice
            results.append(work_request)
            work_requests[action_instance.id] = work_request
        else:
            work_request = work_requests[action_instance.id]

        if pipeline_parameters:
            work_request.environment[pipeline_parameters.parameter] = pipeline_parameters.value

        if include_configuration:
            try:
                work_request.configuration = action_instance.configuration.configuration
            except (TypeError, AttributeError):
                pass

    def get_workable_work_requests(self):
        results = []
        for session in get_db_session():
            app_configuration = session.query(AppConfiguration).get(1)
            if not app_configuration or app_configuration.process_queue:
                work_requests = {}
                self._get_ready_work_requests(session, work_requests, results)
                self._get_stalled_work_requests(session, work_requests, results)

        return results

    def get_verify_working(self, time_difference):
        results = []
        for session in get_db_session():
            for action_instance in session.query(ActionInstance).join(PipelineInstance) \
                    .filter(PipelineInstance.status_id == StatusConstants.INPROGRESS) \
                    .filter(ActionInstance.status_id == StatusConstants.INPROGRESS) \
                    .filter(ActionInstance.assigned_to != '') \
                    .filter(ActionInstance.start_date <= datetime.datetime.utcnow() - datetime.timedelta(minutes=time_difference)) \
                    .filter(ActionInstance.end_date.is_(None)) \
                    .filter(ActionInstance.manual == 0).all():
                results.append(action_instance.serialize())
        return results

    def get_work_request_by_action_instance_id(self, action_instance_id):
        for session in get_db_session():
            work_requests = {}
            results = []
            for action_instance, pipeline_parameters in session.query(ActionInstance, PipelineParameters) \
                    .outerjoin(PipelineParameters, PipelineParameters.pipeline_instance_id == ActionInstance.pipeline_instance_id)\
                    .filter(ActionInstance.id == action_instance_id).all():
                self.configure_work_request(action_instance, pipeline_parameters, work_requests, results, include_configuration=False)
            return results[0].__dict__
        return None

    def get_action_instance_by_id(self, _id, session=None):
        """
        Get ActionInstance by ID
        :param _id: long ID
        :param session: Possible db session to use
        :return: Action Instance
        :rtype ActionInstance
        """
        if session is None:
            for new_session in get_db_session():
                return new_session.query(ActionInstance).get(_id).serialize()
        return session.query(ActionInstance).get(_id)

    def complete_action_instance(self, _id, post_data):
        for session in get_db_session():
            action_instance = self.get_action_instance_by_id(_id, session)
            action_instance.end_date = datetime.datetime.utcnow()

            self.store_service.set_completing(_id)
            self._save_statistics(action_instance, session, post_data)
            self._save_status_information(action_instance, session, post_data)
            self.store_service.clear_completing(_id)

            session.commit()

        return True

    def callback_action_instance(self, _id, post_data):
        for session in get_db_session():
            self._save_status(self.get_action_instance_by_id(_id, session), session, post_data, True)
        return True

    def reset_pipeline_instance(self, pipeline_instance_id):
        for session in get_db_session():
            pipeline_instance = session.query(PipelineInstance)\
                                              .options(
                                                  joinedload(PipelineInstance.stage_instances)
                                                  .joinedload(StageInstance.workflow_instances)
                                                  .joinedload(WorkflowInstance.action_instances)).filter(PipelineInstance.id == pipeline_instance_id).first()

            instance_workflow_engine = InstanceWorkflowEngine(self.status_dal, pipeline_instance)
            instance_workflow_engine.reset_pipeline()
            session.commit()
        return True

    @staticmethod
    def _print_pipeline_instance(pipeline_instance):
        print("pipeline: {}, Status: {}, Start Date: {}, End Date:{}".format(pipeline_instance.id, pipeline_instance.status_id, pipeline_instance.start_date, pipeline_instance.end_date))
        for stage_instance in pipeline_instance.stage_instances:
            print("stage: {}, Status: {}, Start Date: {}, End Date: {}".format(stage_instance.id, stage_instance.status_id, stage_instance.start_date, stage_instance.end_date))
            for workflow_instance in stage_instance.workflow_instances:
                print("  workflow: {}, Status: {}, Start Date: {}, End Date: {}".format(workflow_instance.id, workflow_instance.status_id, workflow_instance.start_date, workflow_instance.end_date))
                for action_instance in workflow_instance.action_instances:
                    print("    action: {}, Status: {}, Start Date: {}, End Date: {}, Order: {}, Slice: {}".format(action_instance.id, action_instance.status_id, action_instance.start_date, action_instance.end_date, action_instance.order, action_instance.slice))

    def reset_action_instance(self, _id, complete_reset=False, check_status=False):  # pylint: disable=unused-argument
        for session in get_db_session():
            action_instance = session.query(ActionInstance).get(_id)
            if check_status and action_instance.status_id != StatusConstants.INPROGRESS:
                return False

            instance_workflow_engine = InstanceWorkflowEngine(self.status_dal, action_instance.pipeline_instance)
            instance_workflow_engine.reset_action(action_instance)

            if self.qa_module is not None:
                self.qa_module.reset_results(action_instance.id, session)
            
            session.commit()
        return True

    def cancel_action_instance(self, action_instance_id):
        for session in get_db_session():
            action_instance = self.get_action_instance_by_id(action_instance_id, session)

            if action_instance:
                serialized = action_instance.serialize()
                instance_workflow_engine = InstanceWorkflowEngine(StatusDal(session), action_instance.pipeline_instance)
                instance_workflow_engine.complete_an_action(action_instance_id, StatusConstants.CANCELED)
                self.queue_constants.cancel_worker(serialized)
                session.commit()
            else:
                raise InvalidObjectException("Action Instance not found", 404)
        return {"message": "Action Instance has been canceled."}

    def reconcile_pipeline_instances(self):
        for session in get_db_session():
            action_query = session.query(ActionInstance).filter(ActionInstance.status_id <= StatusConstants.INPROGRESS).subquery('ai')
            pipeline_instances = session.query(PipelineInstance).filter(PipelineInstance.status_id == StatusConstants.INPROGRESS)\
                .filter(~exists().where(PipelineInstance.id == action_query.c.pipeline_instance_id))
            
            for pipeline_instance in pipeline_instances.all():
                query = session.query(StageInstance, WorkflowInstance, ActionInstance).filter(and_(
                    StageInstance.pipeline_instance_id == pipeline_instance.id,
                    WorkflowInstance.stage_instance_id == StageInstance.id,
                    ActionInstance.workflow_instance_id == WorkflowInstance.id
                )).order_by(asc(StageInstance.id), asc(WorkflowInstance.id), asc(ActionInstance.id))

                highest_status_id = None
                highest_end_date = None
                mappings = {}
                for stage_instance, workflow_instance, action_instance in query.all():
                    if stage_instance.end_date is None:
                        work_name = 'workflow_{}'.format(workflow_instance.id)
                        if work_name not in mappings:
                            mappings[work_name] = {'obj': workflow_instance,
                                                   'highest_status_id': workflow_instance.status_id,
                                                   'highest_end_date': workflow_instance.end_date,
                                                   'stage_obj': stage_instance}

                        if workflow_instance.end_date is None:
                            workflow_check = mappings[work_name]
                            if workflow_check['highest_status_id'] < action_instance.status_id:
                                workflow_check['highest_status_id'] = action_instance.status_id

                            if (workflow_check['highest_end_date'] is None and action_instance.end_date) or \
                               (action_instance.end_date and workflow_check['highest_end_date'] < action_instance.end_date):
                                workflow_check['highest_end_date'] = action_instance.end_date

                for mapping in list(mappings.values()):
                    workflow = mapping['obj']
                    stage = mapping['stage_obj']

                    workflow.status_id = mapping['highest_status_id']
                    workflow.end_date = mapping['highest_end_date']
                    
                    if stage.status_id < workflow.status_id:
                        stage.status_id = workflow.status_id
                        highest_status_id = stage.status_id

                    if (stage.end_date is None and workflow.end_date) or (workflow.end_date and stage.end_date < workflow.end_date):
                        stage.end_date = workflow.end_date
                        highest_end_date = stage.end_date

                if highest_status_id is not None and highest_end_date is not None:
                    logger.info("Reconciling PipelineInstance: {}".format(pipeline_instance.id))
                    pipeline_instance.status_id = highest_status_id
                    pipeline_instance.end_date = highest_end_date

            session.commit()

    def _wait_for_parallel_calculations(self, action_instance):
        is_calculating = StoreService.is_calculating_workflow(action_instance.pipeline_instance_id)
        if is_calculating:
            count = 0
            while count < 5 and is_calculating:
                time.sleep(.1)
                is_calculating = StoreService.is_calculating_workflow(action_instance.pipeline_instance_id)
                count += 1

    def _save_status(self, action_instance, session, post_data, allow_save=False):
        if 'status' in post_data:
            status = self.get_status_by_name(post_data['status'], session)

            if action_instance.callback_required and status.type == StatusTypes.SUCCESS and not allow_save:
                # Allow_save comes from callback_method.
                return

            self._wait_for_parallel_calculations(action_instance)

            pipeline_instance = session.query(PipelineInstance).options(
                joinedload(PipelineInstance.stage_instances)
                .joinedload(StageInstance.workflow_instances)
                .joinedload(WorkflowInstance.action_instances)
            ).options(joinedload(PipelineInstance.parameters)).get(action_instance.pipeline_instance_id)
            workflow_engine = InstanceWorkflowEngine(StatusDal(session), pipeline_instance)
            workflow_engine.complete_an_action(action_instance.id, status.id)
            for instance in workflow_engine.instances_to_add:
                session.add(instance)
            session.commit()

            self.event_service.trigger_possible_event(pipeline_instance, action_instance, session)

    def _save_statistics(self, action_instance, session, post_data):
        try:
            self._save_parameters(action_instance.pipeline_instance_id, session, post_data)
            self._save_stats(action_instance.pipeline_instance_id, session, post_data)
            if self.qa_module is not None:
                self.qa_module.save_results(action_instance, session, post_data)
        except Exception as exception:
            logger.error(exception)

    def _save_status_information(self, action_instance, session, post_data):
        check = '{}__{}'.format(action_instance.pipeline_instance_id, action_instance.action_id)
        if self.store_service.is_completing(check):
            time.sleep(random.randint(3, 10) * 0.01)

        self.store_service.set_completing(check)
        try:
            self._save_status(action_instance, session, post_data)
        except Exception as exception:
            logger.error(exception)
        self.store_service.clear_completing(check)

    def _save_parameters(self, pipeline_instance_id, session, post_data):
        if 'parameters' in post_data:
            parameters_hash = {}
            for pipeline_parameter in session.query(PipelineParameters).filter(PipelineParameters.pipeline_instance_id == pipeline_instance_id).all():
                parameters_hash[pipeline_parameter.parameter] = pipeline_parameter

            for key, value in post_data['parameters'].items():
                pipeline_parameter = None
                if key in parameters_hash:
                    pipeline_parameter = parameters_hash[key]
                else:
                    pipeline_parameter = PipelineParameters(pipeline_instance_id=pipeline_instance_id)
                    session.add(pipeline_parameter)
                pipeline_parameter.parameter = str(key)
                pipeline_parameter.value = str(value)
            session.commit()

    def _save_stats(self, pipeline_instance_id, session, post_data):  # pylint: disable=too-many-locals
        if 'stats' in post_data:
            statistics_cache = {}
            stats_needed = list(post_data['stats'].keys())

            for stat in session.query(Statistics).filter(Statistics.name.in_(list(post_data['stats'].keys()))):
                try:
                    if stat.name in stats_needed:
                        clone_stats = list(stats_needed)
                        for index, item in enumerate(stats_needed):
                            if item == stat.name:
                                del clone_stats[index]
                                break
                        stats_needed = clone_stats
                    statistics_cache[stat.name] = stat
                except Exception:
                    import traceback
                    traceback.print_exc()

            for stat_to_create in stats_needed:
                stat = Statistics(name=stat_to_create)
                session.add(stat)

                statistics_cache[stat.name] = stat

            session.flush()
            session.commit()

            for key, value in post_data['stats'].items():
                stats = PipelineStatistics(pipeline_instance_id=pipeline_instance_id, statistics_id=statistics_cache[key].id, value=value)
                session.add(stats)

            session.commit()

    def partial_edit(self, _id, changes):
        for session in get_db_session():
            return self.edit_object(session, ActionInstance, _id, changes).serialize()

    def get_status_by_name(self, name, session=None):
        if session is not None:
            return session.query(Status).filter(func.lower(Status.name) == name.lower()).first()

        for new_session in get_db_session():
            status = new_session.query(Status).filter(func.lower(Status.name) == name.lower()).first()
            return status.serialize() if status else None



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import logging

try:
    import simplejson as json
except ImportError:
    import json

from functools import wraps
from flask import Response

from rapid.lib.exceptions import HttpException
from rapid.lib.utils import RoutingUtil
from rapid.lib.framework.ioc import IOC

db = None
Base = None

UWSGI = False
try:
    import uwsgi
    UWSGI = True
except ImportError:
    pass


def get_declarative_base():
    global Base  # pylint: disable=global-statement
    if Base is None:
        from sqlalchemy.ext.declarative import declarative_base
        Base = declarative_base()
    return Base


def set_db(_db):
    global db  # pylint: disable=global-statement
    db = _db


def is_primary_worker():
    return uwsgi.worker_id() == 1 if UWSGI else not UWSGI


def get_db_session():
    session = db.session
    try:
        yield session
    finally:
        if session is not None:
            session.rollback()
            session.remove()
            session = None


def setup_config_from_file(app, args):
    if app.rapid_config['_is'] == 'client':
        from ..client import client_configuration
        app.rapid_config = client_configuration.ClientConfiguration(args.config_file)
    elif app.rapid_config['_is'] == 'master':
        from ..master import master_configuration
        app.rapid_config = master_configuration.MasterConfiguration(args.config_file)
    IOC.register_global('rapid_config', app.rapid_config)


def api_key_required(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if 'X-Rapidci-Api-Key' in get_current_request().headers \
                and RoutingUtil.is_valid_request(get_current_request().headers['X-Rapidci-Api-Key'], get_current_app().rapid_config.api_key):
            return func(*args, **kwargs)
        return Response("Not authorized", status=401)
    return decorated_view


def get_current_app():
    from flask import current_app
    return current_app


def get_current_request():
    from flask import request
    return request


def basic_auth(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        config = get_current_app().rapid_config
        try:
            if config.basic_auth_user == get_current_request().authorization['username'] and config.basic_auth_pass == get_current_request().authorization['password']:
                return func(*args, **kwargs)
        except (KeyError, TypeError):
            pass
        return Response("Invalid Authorization", status=401)
    return wrapper


def json_response(exception_class=None, message=None):
    """
    :param exception_class: str
    :param message: str
    """
    def wrap(_f):
        def wrapped_json_response(*args, **kwargs):
            try:
                response = _f(*args, **kwargs)
                return Response(json.dumps(response), content_type="application/json")
            except Exception as exception_stuff:  # pylint: disable=broad-except
                import traceback
                traceback.print_exc()
                if hasattr(exception_stuff, 'get_body'):
                    return exception_stuff
                exception = exception_class(message) if exception_class is not None else Exception(str(exception_stuff))
                response = Response(json.dumps(exception.to_dict())) if hasattr(exception, 'to_dict') else Response(json.dumps({"message": exception.__dict__}))
                response.status_code = exception.status_code if hasattr(exception, 'status_code') else 500
                response.content_type = 'application/json'
                return response
        return wrapped_json_response
    return wrap


def setup_logging(flask_app):
    handler = logging.StreamHandler()
    if hasattr(flask_app.rapid_config, 'log_file') and flask_app.rapid_config.log_file:
        handler = logging.FileHandler(flask_app.rapid_config.log_file)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    # flask_app.logger.addHandler(handler)
    # flask_app.logger.setLevel(logging.INFO)

    logger = logging.getLogger('rapid')
    logger.addHandler(handler)
    logger.propagate = False  # Turn off double logging.
    logger.setLevel(logging.INFO)


def setup_status_route(flask_app):
    @flask_app.route('/status')
    def status():  # pylint: disable=unused-variable
        return 'Running'



from rapid.lib.framework.injectable import Injectable


class QueueHandlerConstants(Injectable):
    __injectables__ = {}
    queue_handler_classes = []

    def __init__(self):
        self.queue_handlers = []
        self.load_handlers()

    def load_handlers(self):
        from rapid.lib import IOC
        for handler_class in self.queue_handler_classes:
            self.queue_handlers.append(IOC.get_class_instance(handler_class))

    def cancel_worker(self, action_instance):  # type: (dict) -> bool
        for handler in self.queue_handlers:
            if handler.can_process_action_instance(action_instance):
                return handler.cancel_worker(action_instance)
        return False



from flask import send_from_directory, current_app

from rapid.lib import api_key_required, basic_auth


class FilesController(object):
    def __init__(self, flask_app):
        self.flask_app = flask_app

    def configure_routing(self):
        self.flask_app.add_url_rule('/get_file/<path:file_path>', 'get_file', basic_auth(self.get_file), methods=['GET'])

    def get_file(self, file_path):
        file_name = "/".join([self.flask_app.rapid_config.static_file_directory, file_path])
        return send_from_directory('/'.join(file_name.split('/')[0:-1]), file_name.split('/')[-1])
